#include <glib.h>
#include "main.hh"

GString *output = g_string_new("");


inline gboolean vertical_black(unsigned char* framebuffer,gint line,gint width,gint height) {
    /* Return TRUE if line is black, else FALSE */
    glong pixel;
    gulong base;
    guint8 R,G,B;

    for (pixel=0;pixel<height;pixel++) {
        base = (line+pixel*width)*3;
        B = *(framebuffer+base);
        G = *(framebuffer+base+1);
        R = *(framebuffer+base+2);
        if ((B<100) && (R<100) && (G<100)) {
            /* Dark pixel */
        } else {
            return FALSE;
        }
    }
    return TRUE;
}

inline gboolean horizontal_black(unsigned char* framebuffer,gint line,gint width) {
    /* Return TRUE if line is black, else FALSE */
    glong pixel;
    gulong base = line*width;
    guint8 R,G,B;

    for (pixel=0;pixel<width;pixel++) {
        B = *(framebuffer+(base+pixel)*3);
        G = *(framebuffer+(base+pixel)*3+1);
        R = *(framebuffer+(base+pixel)*3+2);
        if ((B<100) && (R<100) && (G<100)) {
            /* Dark pixel */
        } else {
            return FALSE;
        }
    }
    return TRUE;
}

/* Check amount of clipping for top, bottom, left and right. There're 2 samples
   taken, to reduce to possibility a whole black frame is selected */
gboolean autoclipper(gchar* filename, gulong frames) {
    mpeg3_t* file;
    gint vstrm,vistream = 0;
    guint i;
    gulong framenr;
    gint cliptop30 = 10000,clipbottom30 = 10000,clipleft30 = 10000,clipright30 = 10000;
    gint cliptop70 = 10000,clipbottom70 = 10000,clipleft70 = 10000,clipright70 = 10000;
    gboolean top,bottom,right,left;
    guint width,height;
    gfloat perc;
    GString *text = g_string_new("");

    /* init random generator */
    srand(time(0));

    /* Open input MPEG2 stream */
    //g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Autoclipper about to open input %s",filename);
    file = mpeg3_open(filename);

    if (file==NULL) {
        /* Error, not able to open MPEG2 stream properly */
        g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"Unable to open %s, possibly not an MPEG2 or IFO file?",filename);
        return FALSE;
    } else {
        g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Autoclipper input %s opened",filename);
    }
    /* Open video stream */
    vstrm=mpeg3_total_vstreams(file);
    vistream=(vistream>vstrm - 1 ? vstrm - 1 : vistream);
    /* Parse parameters of stream */
    width=mpeg3_video_width(file, vistream);
    height=mpeg3_video_height(file, vistream);

    /* Allocate buffers */
    guchar* framebuffer = new guchar[width*height*BYTES_PER_COLOUR_OUTPOUT];
    guchar* extrabuffer = new guchar[width*BYTES_PER_COLOUR_OUTPOUT+4];
    guchar** rowptr = new (guchar*)[height];
    for (i=0; i<height; i++) {
        if(i==height-1) {
            rowptr[i]=extrabuffer;
        } else {
            rowptr[i]=framebuffer+(width*(height-1)*BYTES_PER_COLOUR_OUTPOUT)-(width*BYTES_PER_COLOUR_OUTPOUT*i);
        }
    }

    /* DETERMINE TOP\BOTTOM\LEFT\RIGHT CLIPPING FOR A RANDOM FRAME */
    do {
        perc = random();
        perc = perc/(gfloat)RAND_MAX;
        perc = perc * (gfloat)(frames-40)/(gfloat)frames;
    } while (perc<0.01 || perc>0.99); // ???

    g_string_sprintf(text,"Autoclipper sample offset at %g\%",perc);
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,text->str);
    framenr = mpeg3_seek_percentage(file, perc); 
    /* Read bunch of frames to reach a full one, hope we passed a keyframe */
    for (i=0;i<31;i++) {
        mpeg3_read_frame(file, rowptr, 0, 0, width, height, width, height, LIBMPEG_COLOUR_DEPT_OUPUT, vistream);
    }
    /* Read frame sample */
    if(mpeg3_read_frame(file, rowptr, 0, 0, width, height, width, height, LIBMPEG_COLOUR_DEPT_OUPUT, vistream)) {
        g_log(DRIP_LD,G_LOG_LEVEL_ERROR,"Autoclipper couldn't open frame");
        return FALSE;
    }
    /* Top\Bottom clipping */
    top = TRUE;bottom = TRUE;
    for (i=0;i<(height/2);i++) {
        /* Check lines */
        if (bottom==TRUE)    { bottom = horizontal_black(framebuffer,i,width);}
        if (top==TRUE)       { top = horizontal_black(framebuffer,height-1-i,width);}
        /* Non black line found -> set clipping */
        if (bottom==FALSE)   { clipbottom30 = i+1;bottom=2;}
        if (top==FALSE)      { cliptop30 = i;top=2;}
        /* Ready? */
        if ((top==FALSE) && (bottom==FALSE)) {
            /* Found both lines that are not 100% black */
            i = height;
        }
    }
    /* Left\Right clipping */
    left = TRUE;right = TRUE;
    for (i=0;i<(width/2);i++) {
        /* Check lines */
        if (left==TRUE)   { left = vertical_black(framebuffer,i,width,height);}
        if (right==TRUE)  { right = vertical_black(framebuffer,width-1-i,width,height);}
        /* Non black line found -> set clipping */
        if (left==FALSE)  { clipleft30 = i+1;left=2;}
        if (right==FALSE) { clipright30 = i;right=2;}
        /* Ready? */
        if ((left==FALSE) && (right==FALSE)) {
            /* Found both lines that are not 100% black */
            i = width;
        }
    }

    /* DETERMINE TOP\BOTTOM\LEFT\RIGHT CLIPPING FOR (1-random) FRAME */
    perc = 1-perc;
    while ((frames-(frames*perc))<40) {perc=perc-0.01;}
    g_string_sprintf(text,"Autoclipper shadow sample offset at %g\%",perc);
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,text->str);
    framenr = mpeg3_seek_percentage(file,perc);
    /* Read bunch of frames to reach a full one, hope we passed a keyframe */
    for (i=0;i<31;i++) {
        mpeg3_read_frame(file, rowptr, 0, 0, width, height, width, height, LIBMPEG_COLOUR_DEPT_OUPUT, vistream);
    }
    if(mpeg3_read_frame(file, rowptr, 0, 0, width, height, width, height, LIBMPEG_COLOUR_DEPT_OUPUT, vistream)) {
        g_log(DRIP_LD,G_LOG_LEVEL_ERROR,"Autoclipper couldn't open frame");
        return FALSE;
    }
    /* Top\Bottom clipping */
    top = TRUE;bottom = TRUE;
    for (i=0;i<(height/2);i++) {
        /* Check lines */
        if (bottom==TRUE)    { bottom = horizontal_black(framebuffer,i,width);}
        if (top==TRUE)       { top = horizontal_black(framebuffer,height-1-i,width);}
        /* Non black line found -> set clipping */
        if (bottom==FALSE)    { clipbottom70 = i+1;bottom=2;}
        if (top==FALSE) { cliptop70 = i;top=2;}
        /* Ready? */
        if ((top==FALSE) && (bottom==FALSE)) {
            /* Found both lines that are not 100% black */
            i = height;
        }
    }
    /* Left\Right clipping */
    left = TRUE;right = TRUE;
    for (i=0;i<(width/2);i++) {
        /* Check lines */
        if (left==TRUE)   { left = vertical_black(framebuffer,i,width,height);}
        if (right==TRUE)  { right = vertical_black(framebuffer,width-1-i,width,height);}
        /* Non black line found -> set clipping */
        if (left==FALSE)  { clipleft70 = i+1;left=2;}
        if (right==FALSE) { clipright70 = i;right=2;}
        /* Ready? */
        if ((left==FALSE) && (right==FALSE)) {
            /* Found both lines that are not 100% black */
            i = width;
        }
    }

    /* Fill in global clip settings */
    if (cliptop30>cliptop70) { cliptop = cliptop70;}
        else { cliptop = cliptop30;}
    if (clipbottom30>clipbottom70) { clipbottom = clipbottom70;}
        else { clipbottom = clipbottom30;}
    if (clipleft30>clipleft70) { clipleft = clipleft70;}
        else { clipleft = clipleft30;}
    if (clipright30>clipright70) { clipright = clipright70;}
        else { clipright = clipright30;}
    clipleft+=1; //adjust to real clip.. TODO: better logic for this ;) 
    cliptop+=1; //adjust to real clip.. TODO: better logic for this ;)

    g_string_sprintf(output,"Auto clipping sample top %i,bottom %i,left %i,right %i lines",cliptop,clipbottom,clipleft,clipright);
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,output->str);


    /* Clean & Exit */
    mpeg3_close(file);
    return TRUE;
}


