// Jarl van Katwijk - 2001
// original code: mpeg2divx 0.8
// 2000/10/23 Ulrich Hecht <uli@emulinks.de>
// licensed under the GNU Public License (GPL) v2

// added second audio stream - Rainer Lay 2001

// avi encoder
#include <videoencoder.h>
#include <avifile.h>
#include <aviplay.h>
#include <version.h>
#if (AVIFILE_MAJOR_VERSION == 0) && (AVIFILE_MINOR_VERSION==6) 
   #include <default.h>
   #include <fourcc.h>
   #include <creators.h>
#else
   #include <aviutil.h>
#endif
// system 
#include <sys/time.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <glib.h>
#include <gtk/gtk.h>
#include <unistd.h>
#include <fstream>
#include <errno.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
// mpeg3 decoder
#include <libmpeg3.h>
#include <mpeg3protos.h>
// drip general
#include "../src/drip.h"
#include "../src/external.h"
// filters
#include <spu.h>
#include <subpictures.h>
#include "deinterlace.hh"
#include "autoclipper.hh"
#include "pulldown.hh"
// main
#include "main.hh"
#ifndef STANDALONE
#endif

using namespace std;

// maximum number of audio streams
#define AMAX 2

typedef struct {
    gint in_stream;
    fourcc_t codec;
    gfloat mp3bitrate;
    int achan;
    int rate;
    long samples;
    gfloat sampsperframe_add;
    gint sampsperframe;
    gfloat samplepos;
    gfloat samplepos_last;
    gshort* audiobufferint;
    gshort* read_buffer;
    WAVEFORMATEX fmt;
    IAviAudioWriteStream* outstream;
    gfloat volume_adjustment;
} au_t;
 
typedef struct {
    gint rc_reaction_period;
    gint rc_reaction_ratio;
    gint rc_period;
    gint max_key_interval;
    gint obmc;
    gint quality;
    gint max_quantizer;
    gint min_quantizer;
} opendivx_params_t;

gint clipleft = 0;
gint clipright = 0;
gint cliptop = 0;
gint clipbottom = 0;

#if (AVIFILE_MAJOR_VERSION == 0) && (AVIFILE_MINOR_VERSION==6) 
/* Wrapper for SetAttribute of avifile codec object */
void Codec_SetAttribute(const CodecInfo* codecInfo,const gchar* attribute, gint value) {
    gint result;
    GString *output = g_string_new("");

    /* Try to set attribute */
    result = Creators::SetCodecAttr(*codecInfo, attribute, value);

    /* Handle negative sets */
    if (result==-1) {
        g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"Could not set attribute %s to value %i. Ignoring, wrong codec or outdated avifile?",attribute,value);
    }

    g_string_free(output,TRUE);
    return;
}
#endif

/* Decode Next Frame */
gboolean read_frame(mpeg3_t* file,guchar** rowptr,gint in_x,gint in_y,gint in_w, gint in_h,gint out_w,gint out_h,gint color_model,gint stream) {
#ifndef YCRCB
    return mpeg3_read_frame(file, rowptr, in_x, in_y, in_w, in_h, out_w, out_h, color_model, stream);
#else
    //TODO : finish
    return mpeg3_read_yuvframe(file, Ybuffer,CRbuffer,CBbuffer, in_x, in_y, in_h, in_w, stream);
#endif
}

/* Scale frame */
void scale_frame(gboolean hyper_zoom,GdkPixbuf* source,GdkPixbuf *dest,gint dest_x,gint dest_y,gint dest_width,gint dest_height,double offset_x,double offset_y,double scale_x,double scale_y) {
    if(hyper_zoom==TRUE) {
        gdk_pixbuf_scale(source, dest, dest_x, dest_y, dest_width, dest_height,offset_x, offset_y, scale_x,scale_y,GDK_INTERP_HYPER);
    } else {
        gdk_pixbuf_scale(source, dest, dest_x, dest_y, dest_width, dest_height,offset_x, offset_y, scale_x,scale_y,GDK_INTERP_BILINEAR);
    }
    return;
}

/* Encode frame */
void encode_frame(IAviVideoWriteStream *stream,CImage *frame) {
    stream->AddFrame(frame);
    return;
}

/*
 * Main encoding function
 */
gpointer encode(gpointer data) {
    arg_parmS *arg_parm = (arg_parmS*)data;
    int argc = arg_parm->argc;
    gchar **argv = arg_parm->argv;
    mpeg3_t* file;
    gint result;
    GString *output = g_string_new("");
    gboolean pulldown = FALSE;

    /* Renice */
    // in the non-CORBA version; arg_parm->nice is not set
    // so check it first
    if ((arg_parm->nice<=20) && (arg_parm->nice>=20)) {
        result = nice(arg_parm->nice);
        if (result!=0) {
            g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"Renice failed, used an invalid value");
        }
    }

    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Starting encoding engine...");
    if(argc < 4) {
        printf("\nUsage: %s [OPTION]... [input MPEG]... [output AVI]\n",argv[0]);
        printf("\nOptions:\n");
        printf("  -b\tDivX ;-) bitrate (default 1200)\n");
        printf("  -f\tuse fast-motion codec (default low-motion)\n");
#if (AVIFILE_MAJOR_VERSION == 0) && (AVIFILE_MINOR_VERSION==6) 
        printf("  -d\tuse DIVX4 encoding codec\n");
        printf("  -dq\tDIVX4 quality\n");
        printf("  -dminq\tDIVX4 min quantizer\n");
        printf("  -dmaxq\tDIVX4 max quantizer\n");
        printf("  -c\tuse avifile video codec (for use with Drip GUI)\n");
        printf("  -au\tuse avifile audio codec (for use with Drip GUI)\n");
#endif
        if (AMAX>1)
            printf("  -au2\tuse avifile audio codec (for use with Drip GUI) for audio channel 2\n");
        printf("  -a\tMP3 bitrate (kBit, default auto)\n");
        if (AMAX>1)
           printf("  -a2\tMP3 bitrate (kBit, default auto) for audio channel 2\n");
        printf("  -av\tAudio volume adjustment multiplier (default = 2.0)\n");
        if (AMAX>1)
           printf("  -av2\tAudio volume adjustment multiplier for audio channel 2 (default = 4.0)\n");
        printf("  -e\tnumber of frames to encode (default all)\n");
        printf("  -ct\tnumber of lines to clip from top (default none)\n");
        printf("  -cb\tnumber of lines to clip from bottom (default none)\n");
        printf("  -cl\tnumber of lines to clip from left (default none)\n");
        printf("  -cr\tnumber of lines to clip from right (default none)\n");
        printf("  -ac\tauto clip (default off, disables other clipping options)\n");
        printf("  -bf\tradius for blur filter (default off)\n");
        printf("  -w\toutput width (-1 = input width)\n");
        printf("  -h\toutput height (-1 = auto, when width=-1 too both are same as input)\n");
        printf("  -z\thyper zooming, looks better but very slow (default off)\n");
        printf("  -s\tadd subpictures to result, also needs a CLUT file\n");
        printf("  -r\taspect ratio\n");
        printf("  -ip\tdeinterlace PAL video stream (default off)\n");
        printf("  -in\tdeinterlace (reverse 3:2 pulldown) NTSC video stream (default off) !! Not yet working !!\n");
        printf("  -n\tnum cpus for decoding (default 1)\n");
        printf("  -as\taudio stream number (default 0)\n");
        if (AMAX>1)
            printf("  -as2\taudio stream number (default disabled) for audio channel 2\n");
        //always disabled for now, not enough speed gain to compensate quality loss: printf("  -m\tuse mmx when availeble. Know to be buggish\n");
        printf("  -vs\tvideo stream number (default 0)\n");
        printf("  -l\tlimit output to size in kb, open new output when limit is reached\n");
        printf("\nNOTE: input files should be standalone MPEG2 files, dvd .VOB files\n");
        printf("      and cached drip files are not valid mpeg2 files and must be read\n");
        printf("      in concationation mode: file1.vob:file2.vob:file3.vob\n\n");
        printf("\n" APPNAME " " VERSION " by Jarl van Katwijk <jarl@xs4all.nl>, based on work by Ulrich Hecht <uli@emulinks.de>\n");
        exit(0);
    }

    g_print("Started encoding engine, pid %i\n",getpid());

    gint argbase=0;
    gint divxbitrate=1200;
    gfloat mp3bitrate=-1;
    fourcc_t video_codec=fccDIV3;
    gint endframe=-1;
    gboolean autoclip = FALSE;
    gboolean hyper_zoom = FALSE;
    gboolean blur=FALSE;
    gboolean deinterlace=FALSE;
    gint owidth=-1;
    gint oheight=-1;
    gint numcpus=1;
    au_t au[AMAX];
    for (gint anum=0;anum<AMAX;anum++) {
        au[anum].in_stream = -1;
        au[anum].codec = 0x55;   // Default to MP3 audio 
        au[anum].mp3bitrate = -1;
        au[anum].achan = 0;
        au[anum].rate = 0;
        au[anum].samples = 0;
        au[anum].outstream = NULL;
        au[anum].volume_adjustment = 2.0; 
    }
    /* Set audio channel 0 to default */
    au[0].in_stream = 0;
 
    opendivx_params_t opendivx_params;
    opendivx_params.rc_period=0;
    opendivx_params.rc_reaction_period=0;
    opendivx_params.rc_reaction_ratio=0;
    opendivx_params.max_key_interval=0;
    opendivx_params.obmc=0;
    opendivx_params.quality=5;
    opendivx_params.max_quantizer=12;
    opendivx_params.min_quantizer=2;

    gint austream=0;
    gint vistream=0;
    gint subpicture = -1;
    gchar* clutfile = NULL;
    //static struct spu_img_struct spu;
    gint mmx=1; //default disabled: very little speed gain, and mxx decoding is said to be lossy.
    gint limit=0;
    gfloat aspect = 0;
    gboolean scale = FALSE;

    gboolean moreopts=true;
    gint max_size = 731906048;
    while(moreopts) {
        moreopts=FALSE;

        if(strcmp(argv[argbase+1],"-cl")==0) {
            clipleft=atoi(argv[argbase+2]);
            argbase+=2;
            moreopts=TRUE;
        }
        if(strcmp(argv[argbase+1],"-cr")==0) {
            clipright=atoi(argv[argbase+2]);
            argbase+=2;
            moreopts=TRUE;
        }
        if(strcmp(argv[argbase+1],"-b")==0) {
            divxbitrate=atoi(argv[argbase+2]);
            argbase+=2;
            moreopts=TRUE;
        }
        if(strcmp(argv[argbase+1],"-f")==0) {
            video_codec=fccDIV4;
            argbase++;
            moreopts=TRUE;
        }
#if (AVIFILE_MAJOR_VERSION == 0) && (AVIFILE_MINOR_VERSION==6) 
        if(strcmp(argv[argbase+1],"-d")==0) {
            video_codec=fccDIVX;//mmioFOURCC('d', 'i', 'v', '3');//1482049860;
            argbase++;
            moreopts=TRUE;
        }
#endif
        if(strcmp(argv[argbase+1],"-dq")==0) {
            opendivx_params.quality=atoi(argv[argbase+2]);
            argbase++;
            moreopts=TRUE;
        }
        if(strcmp(argv[argbase+1],"-dminq")==0) {
            opendivx_params.min_quantizer=atoi(argv[argbase+2]);
            argbase++;
            moreopts=TRUE;
        }
        if(strcmp(argv[argbase+1],"-dmaxq")==0) {
            opendivx_params.max_quantizer=atoi(argv[argbase+2]);
            argbase++;
            moreopts=TRUE;
        }
        if(strcmp(argv[argbase+1],"-a")==0) {
            au[0].mp3bitrate=atoi(argv[argbase+2]);
            argbase+=2;
            moreopts=TRUE;
        }
        if(strcmp(argv[argbase+1],"-a2")==0) {
            au[1].mp3bitrate=atoi(argv[argbase+2]);
            argbase+=2;
            moreopts=TRUE;
        }
        if(strcmp(argv[argbase+1],"-av")==0) {
            au[0].volume_adjustment=atof(argv[argbase+2]);
            argbase+=2;
            moreopts=TRUE;
        }
        if(strcmp(argv[argbase+1],"-av2")==0) {
            au[1].volume_adjustment=atof(argv[argbase+2]);
            argbase+=2;
            moreopts=TRUE;
        }
        if(strcmp(argv[argbase+1],"-r")==0) {
            aspect=atof(argv[argbase+2]);
            argbase+=2;
            moreopts=TRUE;
        }
        if(strcmp(argv[argbase+1],"-e")==0) {
            endframe=atoi(argv[argbase+2]);
            argbase+=2;
            moreopts=TRUE;
        }
        if(strcmp(argv[argbase+1],"-c")==0) {
            video_codec=atoi(argv[argbase+2]);
            argbase+=2;
            moreopts=TRUE;
        }
        if(strcmp(argv[argbase+1],"-au")==0) {
            au[0].codec=atoi(argv[argbase+2]);
            argbase+=2;
            moreopts=TRUE;
        }
        if (AMAX>1)
            if(strcmp(argv[argbase+1],"-au2")==0) {
            au[1].codec=atoi(argv[argbase+2]);
            argbase+=2;
            moreopts=TRUE;
        }
        if(strcmp(argv[argbase+1],"-ct")==0) {
            cliptop=atoi(argv[argbase+2]);
            argbase+=2;
            moreopts=true;
        }
        if(strcmp(argv[argbase+1],"-cb")==0) {
            clipbottom=atoi(argv[argbase+2]);
            argbase+=2;
            moreopts=true;
        }
        if(strcmp(argv[argbase+1],"-z")==0) {
            hyper_zoom = TRUE;
            argbase+=1;
            moreopts=true;
        }
        if(strcmp(argv[argbase+1],"-ip")==0) {
            deinterlace = TRUE;
            argbase+=1;
            moreopts=true;
            g_log(DRIP_LD,G_LOG_LEVEL_INFO,"Activated PAL deinterlace filter");
        }
        if(strcmp(argv[argbase+1],"-in")==0) {
            pulldown = TRUE;
            argbase+=1;
            moreopts=true;
            g_log(DRIP_LD,G_LOG_LEVEL_INFO,"Activated NTSC reverse 3:2 pulldown. NOT YET READY!");
        }
        if(strcmp(argv[argbase+1],"-ac")==0) {
            autoclip = TRUE;
            argbase+=1;
            moreopts=true;
        }
        if(strcmp(argv[argbase+1],"-bf")==0) {
            blur=atoi(argv[argbase+2]);
            argbase+=2;
            moreopts=true;
        }
        if(strcmp(argv[argbase+1],"-w")==0) {
            owidth=atoi(argv[argbase+2]);
            scale = TRUE;
            argbase+=2;
            moreopts=true;
        }
        if(strcmp(argv[argbase+1],"-h")==0) {
            oheight=atoi(argv[argbase+2]);
            scale = TRUE;
            argbase+=2;
            moreopts=true;
        }
        if(strcmp(argv[argbase+1],"-as")==0) {
            au[0].in_stream=atoi(argv[argbase+2]);
            argbase+=2;
            moreopts=true;
        }
        if (AMAX>1) {
            if(strcmp(argv[argbase+1],"-as2")==0) {
                au[1].in_stream=atoi(argv[argbase+2]);
                argbase+=2;
                moreopts=true;
            }
        }
        if(strcmp(argv[argbase+1],"-vs")==0) {
            vistream=atoi(argv[argbase+2]);
            argbase+=2;
            moreopts=true;
        }
        if(strcmp(argv[argbase+1],"-s")==0) {
            subpicture=atoi(argv[argbase+2]);
            clutfile= argv[argbase+3];
            argbase+=3;
            moreopts=true;
        }
        if(strcmp(argv[argbase+1],"-n")==0) {
            numcpus=atoi(argv[argbase+2]);
            argbase+=2;
            moreopts=true;
        }
        if(strcmp(argv[argbase+1],"-m")==0) {
            mmx=1;
            argbase+=1;
            moreopts=true;
        }
        if(strcmp(argv[argbase+1],"-l")==0) {
            limit=1;
            max_size=atoi(argv[argbase+2]);
            argbase+=2;
          moreopts=true;
        }

        if(strcmp(argv[argbase+1],"-A")==0) {
            argbase+=1;
            moreopts=true;
        }
        if(strcmp(argv[argbase+1],"-F")==0) {
            argbase+=1;
            moreopts=true;
        }
    }

    /* Chunk size */
    if (max_size<50000) {
        /* Stop polling loops */
        encoding = TRUE;
        usleep(15000);
#ifndef STANDALONE
        gui_ready();
#endif
        encoding = FALSE;
        g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"Max size too small, use at least 50000kb");
        return 0;
    } else {
        /* Adjust max size so max_size_new + header =~ max_size */
        max_size=max_size-50000;
    }

    /* Input files */
    int numfiles=argc-argbase-2;
    GString* inputfiles[numfiles];
    for(int i=argbase+1; i<argc-1; i++) {
        inputfiles[i-argbase-1] = g_string_new(argv[i]);
        g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Input %d out %s",i-argbase-1,inputfiles[i-argbase-1]->str);
    }
    char* outputfile=argv[argc-1];

    /* Open mpeg2 stream (video+audio) */
    g_log(DRIP_LD,G_LOG_LEVEL_MESSAGE,"Opening %s",inputfiles[0]->str);
    file = mpeg3_open(inputfiles[0]->str);
    if (file==NULL) {
        /* Error, not able to open stream properly */
        g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"Unable to open %s, possibly not a plain MPEG2 or valid IFO file",inputfiles[0]->str);
        encoding = TRUE;
#ifndef STANDALONE
        gui_ready();
#endif
        encoding = FALSE;
        return 0;
    }

    /* Open audio stream */
    gint astrm=mpeg3_total_astreams(file);
    gboolean audio_coding=FALSE;
    for (gint anum=0;anum<AMAX;anum++) {
        /* Set audio stream */
        au[anum].in_stream=(au[anum].in_stream>astrm - 1 ? astrm - 1 : au[anum].in_stream);
        if (au[anum].in_stream >= 0) {
            au[anum].achan=mpeg3_audio_channels(file, au[anum].in_stream);
            au[anum].rate=mpeg3_sample_rate(file, au[anum].in_stream);
            au[anum].samples=mpeg3_audio_samples(file, au[anum].in_stream);
            audio_coding=TRUE;
            g_log(DRIP_LD,G_LOG_LEVEL_INFO,"Audio %d: Using audio stream %d of %d, Channels %d, rate %d, samples %li",anum, au[anum].in_stream, astrm, au[anum].achan,au[anum].rate,au[anum].samples);
            g_log(DRIP_LD,G_LOG_LEVEL_INFO,"Audio %d bitrate %g",anum,au[anum].mp3bitrate);
            g_log(DRIP_LD,G_LOG_LEVEL_INFO,"Audio %d volume multiplier %g",anum,au[anum].volume_adjustment);
        }
    }
   
    /* Open video stream */
    int vstrm=mpeg3_total_vstreams(file);
    vistream=(vistream>vstrm - 1 ? vstrm - 1 : vistream);
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Using video stream %d of %d", vistream, vstrm);

    /* Audio channels */
    if (! audio_coding)
        g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"There will be NO audio encoded!");

    /* Video input is a IFO file? -> precalculate total frames */
    ulong vobtotalframes = 0;
    GString *extension = g_string_new((gchar*)(inputfiles[0]->str + inputfiles[0]->len - 3));
    if (strcasecmp("IFO",extension->str)==0) {
         /* Got an IFO :
          *  - open 1st VOB,
          *  - get estimated frames
          *  - calculate total frames
          */
         GString *sequence = g_string_new((gchar*)(inputfiles[0]->str + inputfiles[0]->len - 8));
         sequence = g_string_truncate(sequence,2);
         GString *dir = g_string_new(inputfiles[0]->str);
         dir = g_string_truncate(dir,dir->len - 12);
         GString *vob = g_string_new("");
         if (strcmp(extension->str,"IFO")==0) {
             g_string_sprintf(vob,"%sVTS_%s_1.VOB",dir->str,sequence->str);
         } else {
             g_string_sprintf(vob,"%svts_%s_1.vob",dir->str,sequence->str);
         }
         /* Open VOB */
         mpeg3_t* vobfile = mpeg3_open(vob->str);
         long vobframes   = mpeg3_video_frames(vobfile, vistream);
         struct stat vobstat; stat(vob->str, &vobstat);
         ulong vobtotalsize = vobstat.st_size / 1024;
         float bytes_frame = (float)vobframes / (float)vobtotalsize;

         /* Loop through all VOB files and calculate total frames */
         int i = 2;
         int result;
         do {
             if (strcmp(extension->str,"IFO")==0) {
                 g_string_sprintf(vob,"%sVTS_%s_%i.VOB",dir->str,sequence->str,i);
             } else {
                 g_string_sprintf(vob,"%svts_%s_%i.vob",dir->str,sequence->str,i);
             }
             result = stat(vob->str, &vobstat);
             if (result == 0) {
                 /* Add size to total size */
                 vobtotalsize += vobstat.st_size/1024;
             }
             i++;
         } while (result == 0);

         vobtotalframes = (ulong)(bytes_frame * vobtotalsize);

         /* Clean */
         g_string_free(sequence,TRUE);
         g_string_free(vob,TRUE);
         g_string_free(dir,TRUE);
    }
    g_string_free(extension,TRUE);

    /* SMP */
    if (numcpus>1) {
         mpeg3_set_cpus(file, numcpus);
    } 
   
    /* MMX */
    if (mmx==1) {
         mpeg3_set_mmx(file,TRUE);
    } else {
         mpeg3_set_mmx(file,FALSE);
    }

    struct stat VOB_file;
    ulong total_bytes;
    float frames_per_bytes;
    uint width=mpeg3_video_width(file, vistream);
    uint height=mpeg3_video_height(file, vistream);
    float framerate = mpeg3_frame_rate(file, vistream);
    long frames=mpeg3_video_frames(file, vistream);
    g_log(DRIP_LD,G_LOG_LEVEL_INFO,"Input video reads %dx%d %.2ffps",width,height,framerate);

    /* Setup frames/bytes for more correct total frames estimation */
    stat(inputfiles[0]->str, &VOB_file);
    total_bytes = VOB_file.st_size;
    frames_per_bytes = (float)frames / (float)total_bytes;
    glong todo_frames = frames;
    /* Frames to encode set? -> set vobtotalframes\todo_frames if so */
    if (endframe!=-1) {
        vobtotalframes = endframe;
        todo_frames = endframe;
    } else {
        endframe = frames;
    }

    /* Allocate buffers */
    guchar* framebuffer = new guchar[width*height*BYTES_PER_COLOUR_OUTPOUT];
    guchar* extrabuffer = new guchar[width*BYTES_PER_COLOUR_OUTPOUT+4];
    guchar** rowptr = new (guchar*)[height];
    for (guint i=0; i<height; i++) {
        if(i==height-1) {
            rowptr[i]=extrabuffer;
        } else {
            rowptr[i]=framebuffer+(width*(height-1)*BYTES_PER_COLOUR_OUTPOUT)-(width*BYTES_PER_COLOUR_OUTPOUT*i);
        }
    }
#ifdef YCRCB
    gchar* Ybuffer = new (gchar)[width*height];
    gchar* CRbuffer = new (gchar)[width*height/4];
    gchar* CBbuffer = new (gchar)[width*height/4];
#endif

    /* Audio sample rate & buffers */
    for (int anum=0;anum<AMAX;anum++) {
        if (au[anum].in_stream>=0) {
            au[anum].sampsperframe_add = au[anum].rate/framerate;
            au[anum].samplepos = 0;
            au[anum].samplepos_last = 0;
            au[anum].sampsperframe = (gint)au[anum].sampsperframe_add + 1;
            au[anum].audiobufferint = new gshort[au[anum].sampsperframe*2];
            au[anum].read_buffer = new gshort[au[anum].sampsperframe];
        }
    }

    /* Automatic Clipping */
    if (autoclip==TRUE && endframe<200) {
        g_log(DRIP_LD,G_LOG_LEVEL_INFO,"Autoclipper disbaled because not enough frames (200) in source");
        autoclip = FALSE;
    }
    if (autoclip==TRUE) {
        gint _cliptop = 10000;
        gint _clipbot = 10000;
        gint _cliprig = 10000;
        gint _cliplef = 10000;
        gint i;
        for (i=0;i<7;i++) { /* 7 too high\low?? */
            g_log(DRIP_LD,G_LOG_LEVEL_INFO,"Autoclipper taking sample...");
            autoclipper(inputfiles[0]->str,endframe);
            if (_cliptop>cliptop) { _cliptop = cliptop;}
            if (_clipbot>clipbottom) { _clipbot = clipbottom;}
            if (_cliplef>clipleft) { _cliplef = clipleft;}
            if (_cliprig>clipright) { _cliprig = clipright;}
        }
        cliptop = _cliptop;
        clipbottom = _clipbot;
        clipleft = _cliplef;
        clipright = _cliprig;
    }
    if ((cliptop!=0) || (clipbottom!=0) || (clipleft!=0) || (clipright!=0)) {
        scale = TRUE;
    };
    g_log(DRIP_LD,G_LOG_LEVEL_INFO,"Clipping top %i,bottom %i,left %i,right %i",cliptop,clipbottom,clipleft,clipright);
    g_log(DRIP_LD,G_LOG_LEVEL_INFO,"Input video clipped %ix%i %.2ffps",width-(clipleft+clipright),height-(cliptop+clipbottom),framerate);


    /* Zooming related sizes */
    gboolean correct_aspect = FALSE;
    if(oheight==-1 && owidth==-1) { // All auto
        owidth=width-(clipleft+clipright);
        oheight=height-(cliptop+clipbottom);
        correct_aspect = TRUE;
    }
    if(owidth==-1 && oheight!=-1) { // Width auto
        owidth=width-(clipleft+clipright);
        correct_aspect = TRUE;
    }
    if(oheight==-1 && owidth!=-1) { // Height auto
        oheight = height-(cliptop+clipbottom);
        correct_aspect = TRUE;
    }

    /* Aspect scaling */
    if (correct_aspect==TRUE) {
        gfloat ow = owidth;
        gfloat oh = oheight;
        gfloat as = (ow/oh);
        gfloat correct;
        if (aspect==0) {
            /* No aspect given, correct based on sizes we got so far, no very correct... */
            if (as>2.06388888) {
                /* 2.35:1 aspect */
                aspect = (gfloat)2.35/(gfloat)1;
            }
            if ((as>1.5555555) && (as<2.06388889)) {
                /* 16:9 aspect */
                aspect = (gfloat)16/(gfloat)9;
            }
            if (as<1.5555556) {
                /* 4:3 aspect */
                aspect = (gfloat)4/(gfloat)3;
            }
        }
        g_log(DRIP_LD,G_LOG_LEVEL_INFO,"Using DVD aspect %.2f:1",aspect);
        correct = aspect / ((gfloat)width/(gfloat)height);
        oheight = (gint)(oh / correct);
        oheight = (gint)(oheight / ((gfloat)width/(gfloat)owidth));
    }

    /* Adjust Width & Height to a multipe of 16 for better encoding */
    owidth = ((gint)owidth/16)*16;
    oheight= ((gint)oheight/16)*16;
    g_log(DRIP_LD,G_LOG_LEVEL_INFO,"Output video corrected %ix%i at %.2fFPS",owidth,oheight,framerate);


    /* Output Video header */
    VideoEncoderInfo info;
    memset(&(info.header),0,sizeof(BITMAPINFOHEADER));
    info.header.biSize=sizeof(BITMAPINFOHEADER);
    info.header.biPlanes=1;
    info.header.biBitCount=AVIFILE_COLOUR_DEPT_INPUT;
#ifdef YCRCB
    info.header.biCompression=fccYUV;
#endif
    info.header.biWidth = owidth;
    info.header.biHeight = oheight;
    info.header.biSizeImage=info.header.biWidth*info.header.biHeight*BYTES_PER_COLOUR_OUTPOUT;
    info.compressor = video_codec;
    info.quality = 10000;
    info.keyfreq = 200;

    /* Output codec */
    switch (video_codec) {
#if (AVIFILE_MAJOR_VERSION == 0) && (AVIFILE_MINOR_VERSION==6) 
        case fccDIVX:    g_log(DRIP_LD,G_LOG_LEVEL_INFO,"Encoding to OpenDIVX"); 
                         break;
#endif
        case fccDIV3:    g_log(DRIP_LD,G_LOG_LEVEL_INFO,"Encoding to MPEG4 low motion");
                         break;
        case fccDIV4:    g_log(DRIP_LD,G_LOG_LEVEL_INFO,"Encoding to MPEG4 high motion");
                         break;
        default:         g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"No codec, not encoding.");
                         return NULL;
    }

    /* Output codec attributes */
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"About to initialise avifile, when Drip stops (crashes) here get yourself a working avifile...");
    /* Create output avi stream */
    IAviSegWriteFile* avifile = CreateSegmentedFile(outputfile,max_size,0,00777);
#if (AVIFILE_MAJOR_VERSION == 0) && (AVIFILE_MINOR_VERSION==6) 
    const CodecInfo* codecInfo=CodecInfo::match(video_codec);
    if (video_codec==fccDIV3 || video_codec==fccDIV4) { /* MSMPEG4 */
       Codec_SetAttribute(codecInfo, "Crispness", 100);
       Codec_SetAttribute(codecInfo, "KeyFrames", info.keyfreq);
       Codec_SetAttribute(codecInfo, "BitRate", divxbitrate);
    }
    if (video_codec==fccDIVX) { /* OpenDIVX, DIVX4 */
        Codec_SetAttribute(codecInfo, "BitRate",divxbitrate*1500);
        Codec_SetAttribute(codecInfo, "rc_period",info.keyfreq*10);
        //What's this? : Codec_SetAttribute(codecInfo, "MaxAuto",6);
        Codec_SetAttribute(codecInfo, "rc_reaction_period",(gint)(info.keyfreq/20));
        Codec_SetAttribute(codecInfo, "rc_reaction_ratio",(gint)(info.keyfreq/10));
        //Codec_SetAttribute(codecInfo, "max_key_interval",(gint)((info.keyfreq*3)/2));
        //Codec_SetAttribute(codecInfo, "obmc",0); //TODO
        Codec_SetAttribute(codecInfo, "quality",5); //opendivx_params.quality); // DIVX:5=best,1=lowest, avifile:1=enable VBR... ????
        Codec_SetAttribute(codecInfo, "max_quantizer",opendivx_params.max_quantizer);
        Codec_SetAttribute(codecInfo, "min_quantizer",opendivx_params.min_quantizer);
        //Codec_SetAttribute(codecInfo, "bidirect",0); //TODO:2pass encoding,0 = disabled
        //Codec_SetAttribute(codecInfo, "deinterlace",1); //TODO
    }
#else
    IVideoEncoder::SetExtendedAttr(codec, "BitRate", divxbitrate);
    IVideoEncoder::SetExtendedAttr(codec, "KeyFrames", info.keyfreq);
    IVideoEncoder::SetExtendedAttr(codec, "Crispness", 100);
#endif

    /* Add video stream to output avi */
    IAviVideoWriteStream* stream=avifile->AddVideoStream(info.compressor, &(info.header), (guint)(1000000./framerate));
    stream->SetQuality(info.quality);
    stream->SetKeyFrame(info.keyfreq);
    stream->Start();
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Avifile initialisation calls passed");
    g_log(DRIP_LD,G_LOG_LEVEL_INFO,"Video Encoding bitrate %i",divxbitrate);
    g_log(DRIP_LD,G_LOG_LEVEL_INFO,"DIVX Chunk size %i bytes",max_size);
    g_log(DRIP_LD,G_LOG_LEVEL_INFO,"DIVX Output File %s",outputfile);

    /* Buffers */
    unsigned char* scaledbuffer = new unsigned char[info.header.biHeight*info.header.biWidth*BYTES_PER_COLOUR_OUTPOUT];
    CImage* im;
    if (scale == TRUE) {
        im = new CImage((BitmapInfo*)&(info.header),scaledbuffer, false);
    } else {
        im = new CImage((BitmapInfo*)&(info.header),framebuffer, false);
    }
    guchar* _RGBbuffer = (guchar*)malloc(info.header.biHeight*info.header.biWidth*3);

    /* Scaling buffers */
    GdkPixbuf*  gframe = gdk_pixbuf_new_from_data( framebuffer+clipbottom*width*BYTES_PER_COLOUR_OUTPOUT, GDK_COLORSPACE_RGB, FALSE, 8, width, height-(cliptop+clipbottom), width*BYTES_PER_COLOUR_OUTPOUT, NULL, NULL);
    GdkPixbuf*  gscaled = gdk_pixbuf_new_from_data( scaledbuffer, GDK_COLORSPACE_RGB, FALSE, 8, info.header.biWidth, info.header.biHeight, info.header.biWidth*BYTES_PER_COLOUR_OUTPOUT, NULL, NULL);

    /* Output Audio header */
    for (int anum=0;anum<AMAX;anum++) {
        if (au[anum].in_stream>=0) {
            /* Output Audio header */
            memset(&au[anum].fmt,0,sizeof(WAVEFORMATEX));
            au[anum].fmt.wFormatTag=1;
            au[anum].fmt.nChannels=(au[anum].achan>2 ? 2 : au[anum].achan);
            au[anum].fmt.nSamplesPerSec=au[anum].rate;
            au[anum].fmt.nAvgBytesPerSec=(au[anum].achan>1 ? au[anum].rate : au[anum].rate*2);
            au[anum].fmt.nBlockAlign=(au[anum].achan>1 ? 2 : 1);
            au[anum].fmt.wBitsPerSample=16;
            au[anum].fmt.cbSize=sizeof(WAVEFORMATEX);
       
            if (au[anum].mp3bitrate == -1)
                switch(au[anum].rate) {
                case 48000: au[anum].mp3bitrate=(au[anum].achan>1?160:80*au[anum].achan); break;
                case 44100: au[anum].mp3bitrate=(au[anum].achan>1?112:56*au[anum].achan); break;
                case 22050: au[anum].mp3bitrate=(au[anum].achan>1?64:32*au[anum].achan); break;
                default:    g_log(DRIP_LD,G_LOG_LEVEL_ERROR,"Audio %d: unknown sample rate %d Hz, please specify MP3 bitrate",anum, au[anum].rate);
                            exit(1);
                 }
            if (au[anum].achan>0) {
                au[anum].outstream=avifile->AddAudioStream(au[anum].codec, &au[anum].fmt, (int)(au[anum].mp3bitrate*1000)/8);
                au[anum].outstream->Start();
            }
        }
    }

    glong totalframes=0;
    struct timeval tv;
    struct timezone tz={0,0};
    gettimeofday(&tv,&tz);
    glong startsec=tv.tv_sec;
    glong startusec=tv.tv_usec;
    double fps=0;
    gint total_channels;
    gint assetChannels;
    int16_t *input_ptr;
    int16_t *output_ptr;
    gchar *buffer;
    // audio
    gchar *audio_buffer;
    gint downmix51;

    int16_t *input_end;
    off_t len;
    gint cached_frames = 0;
    glong totalframes_countdown = 0;
    //gint framefile_fd;
    GString *framefile = g_string_new("");
    glong writeout_frames_countdown = 1;
    glong writeout_frames_countdown_amount = 5;
    gpointer deinterlace_buffer;
    g_string_sprintf(framefile,"%s/.drip/frame.raw",getenv("HOME"));

    /* Open subpicture */
    if (subpicture > -1) {
        subpicture_init(subpicture,width,height,(u_int*)framebuffer,framerate,clutfile);
        g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Initialized subpicture %i",subpicture);
    }

    /* Init reverse 3:2 pulldown */
    if (pulldown) {
        pulldown_init(width,height,false,0);  //TODO: odd frame & offset shou;d not be hardcoded
        g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Initialized reverse 3:2 pulldown");
    }


    /* 
     * Encoding starts
     */
    encoding = TRUE;
    for (gint j=0; j< numfiles; j++) {
        if(j>0) {
            /* Close previous input stream */
            mpeg3_close(file);
            /* Open new input stream */
            file   = mpeg3_open(inputfiles[j]->str);
            /* Parse frames */
            frames = mpeg3_video_frames(file, vistream);
            g_log(DRIP_LD,G_LOG_LEVEL_MESSAGE,"Opening input AV stream from %s with %li frames",inputfiles[j]->str,frames);
            if (file==0) {
                /* Error, not able to open input stream properly */
                g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"Unable to open %s, possibly not a plain MPEG2 or valid IFO file",inputfiles[0]->str);
                encoding = TRUE;
#ifndef STANDALONE
                gui_ready();
#endif
                /* Close streams */
                stream->Stop();
                delete avifile;
                mpeg3_close(file);
                /* Encoding done */
                encoding = FALSE;
            }
            /* Reset total frames counter */
            totalframes_countdown = 24;
            /* calculate new estimated total frames */
            stat(inputfiles[j]->str, &VOB_file);
            total_bytes += VOB_file.st_size;
            todo_frames = (long)((float)total_bytes * (float)frames_per_bytes);
        }

        /* Read input frames */
        gint quit = FALSE;
        for (gint i=0;quit==FALSE; i++, totalframes++) {

            /* Reached endframe? */
            quit = (totalframes == (endframe-1));

            /* Paused? */
            while (encoding_paused) {
                usleep(1000000);
            }

            /* Stop? */
            if (stopped) {
                stopped = FALSE;
                //printf("\tstopped.\n");
                todo_frames = totalframes;
                quit = TRUE;
                j = numfiles;
                break;
            }

            /* Notify GUI with progress every ?100? frames */
            if ( (totalframes_countdown--) == 0) {
                ulong todo = 0;
                totalframes_countdown = 24;
                if (todo_frames == 0) {
                    todo = vobtotalframes;
                } else {
                    todo = todo_frames;
                }
#ifndef STANDALONE
                gui_progress(totalframes,todo,j,numfiles);
#else
                g_print("\rEncoding frame %i of %i",totalframes,todo);
#endif
            }

            /* Decode Next Frame */
            if(read_frame(file, rowptr, 0, 0, width, height, width, height, LIBMPEG_COLOUR_DEPT_OUPUT, vistream)) {
                /* Reached end of input stream */
                todo_frames = totalframes;
                quit = TRUE;
                /* Update caching */
                cached_frames = 0;
                break;
            }

       /* Audio handling */
       for (int anum=0;anum<AMAX;anum++) {
         if(au[anum].in_stream>=0) {
           downmix51 = 0;
           result = 0;
           au[anum].samplepos += au[anum].sampsperframe_add;
           au[anum].sampsperframe = (gint)au[anum].samplepos - (gint)au[anum].samplepos_last;
           au[anum].samplepos_last = au[anum].samplepos;
           len = (off_t)au[anum].sampsperframe;
           audio_buffer = (gchar *)au[anum].audiobufferint;
           output_ptr = NULL;
           input_ptr = (int16_t*)au[anum].read_buffer;
           input_end = input_ptr + len;
           assetChannels = au[anum].achan;
           if (assetChannels > 2) {
             assetChannels = 2;
           }
           if(au[anum].achan > assetChannels) {
             int16_t *output_end;
             downmix51 = 1;
             for(gint channel = 0; channel < assetChannels; channel++) {
               output_ptr = (int16_t*)audio_buffer + channel;
               output_end = (int16_t*)output_ptr + len * assetChannels;
               while(output_ptr < output_end) {
                 *output_ptr++ = 0;
               }
             }
           }
           total_channels = (downmix51 ? au[anum].achan : assetChannels);
           for(gint channel = 0; 
               channel < total_channels && !result; 
               channel++) {
             if (channel == 0)
               result = mpeg3_read_audio(file, 0,au[anum].read_buffer,channel,len,au[anum].in_stream);
             else
               result = mpeg3_reread_audio(file,0,au[anum].read_buffer,channel,len,au[anum].in_stream);
 
             /* Interlace the audio for the sound driver */
             if(downmix51) {
               input_ptr = (int16_t*)au[anum].read_buffer;
               input_end = input_ptr + len;
               switch(channel)
                 {
                 case 0: output_ptr = (int16_t*)audio_buffer;       break; // Left
                 case 1: output_ptr = (int16_t*)audio_buffer;       break; // Center
                 case 2: output_ptr = (int16_t*)audio_buffer + 1;   break; // Right
                 case 3: output_ptr = (int16_t*)audio_buffer;       break; // Left2
                 case 4: output_ptr = (int16_t*)audio_buffer + 1;   break; // Right2
                 }
             
               if(channel == 1) 
                 while(input_ptr < input_end) {
                   *output_ptr++ += (gint)((*input_ptr) * au[anum].volume_adjustment);
                   *output_ptr++ += (gint)((*input_ptr++) * au[anum].volume_adjustment);
                 }
               else 
                 while(input_ptr < input_end) {
                   *output_ptr += (gint)((*input_ptr++) * au[anum].volume_adjustment);
                   output_ptr += assetChannels;
                 }
             
             } else {
               output_ptr = (int16_t*)audio_buffer + channel;
               input_ptr = (int16_t*)au[anum].read_buffer;
               input_end = input_ptr + len;
               while(input_ptr < input_end) {
                 *output_ptr = (gint)((*input_ptr++) * au[anum].volume_adjustment);
                 output_ptr += assetChannels;
               }
             }
           }
 
           if(au[anum].achan>1) {
             au[anum].outstream->AddData(au[anum].audiobufferint,au[anum].sampsperframe*2);
           } else {
             au[anum].outstream->AddData(au[anum].audiobufferint,au[anum].sampsperframe);
           }
         }
       }
       

            /* Deinterlace PAL */
            if (deinterlace==TRUE) {
                deinterlace_buffer = deinterlace_apply(framebuffer,width,height,totalframes);
                memcpy(framebuffer,deinterlace_buffer,width*height*3);
            }

            /* Deinterlace NTSC (reverse 3:2 pulldown) */
            /* !!!! TODO : finish, this is incomplete code !!! */
            /* !!!!        Needs audio encoding to work 1st !! */
            if (pulldown==TRUE) {
                unsigned char* tmpbuffer = inverse_pulldown(framebuffer, width, height, totalframes);
                if (tmpbuffer!=framebuffer && tmpbuffer) {
                    memcpy(framebuffer, tmpbuffer, width * height * 3);
                }
            }
                
            /* Blurring */
            if (blur==TRUE) {
                im->Blur(blur);
            }

            /* Subpicture */
            if (subpicture > -1) {
                /* Overlay spubpicture to current frame */
                subpicture_overlay((u_int*)framebuffer,width,height,clipbottom);
            }

            /* Scaling */
            if (scale==TRUE) {
                scale_frame(hyper_zoom,gframe, gscaled, 0, 0, info.header.biWidth, info.header.biHeight,-clipleft*info.header.biWidth / double(width), 0, info.header.biWidth / double(width-(clipleft+clipright)), info.header.biHeight / double(height-(cliptop+clipbottom)));
            }

#ifndef STANDALONE
            /* Writeout current frame every 250 frames */
            //TODO: once working ok speed up 
            if ( (writeout_frames_countdown--) == 0 ) {
                /* Next writeout */
                writeout_frames_countdown = writeout_frames_countdown_amount;
                if (writeout_frames_countdown_amount<500) {
                    writeout_frames_countdown_amount+=5;
                } else {
                    writeout_frames_countdown_amount = 499;
                }

                /* Open frame file */
                //framefile_fd = open(framefile->str,O_WRONLY|O_CREAT|O_TRUNC,S_IREAD|S_IWRITE);
                guchar* BGRbuffer; 
                if (scale == TRUE) {
                    BGRbuffer = scaledbuffer;
                } else {
                    BGRbuffer = framebuffer;
                }
                guchar* RGBbuffer;
                /* Yaxis invert and BGR->RGB convert */
                for (gint y=0;y<info.header.biHeight;y++) {
                    RGBbuffer = _RGBbuffer+(info.header.biHeight-1-y)*(info.header.biWidth*3);
                    for (gint x=0;x<info.header.biWidth;x++) {
                        *RGBbuffer = *(BGRbuffer+2);RGBbuffer++;
                        *RGBbuffer = *(BGRbuffer+1);RGBbuffer++;
                        *RGBbuffer = *BGRbuffer;RGBbuffer++;
                        BGRbuffer+=3; 
                    }
                }
                /* Writeout frame */
                //write(framefile_fd,_RGBbuffer,info.header.biHeight*info.header.biWidth*3);
                //close(framefile_fd);
                gui_frame(_RGBbuffer,totalframes,info.header.biWidth,info.header.biHeight);
            }
#endif

            /* Encode next frame (if not skipping) */
            encode_frame(stream,im);
        }
    }

    /* Stop & close encoding */
    stream->Stop();
    //buggish, not needed: astream->Stop();
    delete avifile;
    mpeg3_close(file);
    /* Display encoding info */
    gettimeofday(&tv,&tz);
    fps=totalframes/((tv.tv_sec+tv.tv_usec/1000000.0)-(startsec+startusec/1000000.0));
    g_log(DRIP_LD,G_LOG_LEVEL_INFO,"Finished, %li frames in %g fps",totalframes,fps);
    g_string_free(output,TRUE);
    /* Notify GUI encoding has finished */
#ifndef STANDALONE
    gui_ready();
#endif
    /* Clean & Exit */
    if (pulldown==TRUE) pulldown_cleanup();
    free(_RGBbuffer);
    delete(framebuffer);
    gdk_pixbuf_finalize(gframe);
    delete(scaledbuffer);
    gdk_pixbuf_finalize(gscaled);
    delete(extrabuffer);
    delete(rowptr);
    /* Encoding done */
    encoding = FALSE;
#ifdef STANDALONE
    return NULL;
#else
    pthread_exit(0);
#endif
}

