// Jarl van Katwijk - 2001
// licensed under the GNU Public License (GPL) v2

#include "main.hh"
#include "encoder.hh"
#ifndef STANDALONE
#include "../src/external.h"
#endif
#include <signal.h>
#include <setjmp.h>

#include <ostream.h>
#include <fstream>
#include <strstream>
#include <cstdlib>
#include <cstdio>
#include <string.h>
#include <dirent.h>
#include <dlfcn.h>
// #include <string>


using std::cout;
using std::cerr;
using std::endl;

gboolean encoding = FALSE;
gboolean stopped = FALSE;
gboolean encoding_paused = FALSE;
pthread_t serverthread;
gint serverthread_id;
pthread_t encoderthread;
gint encoderthread_id;
pthread_t guiclientthread;
gint guiclientthread_id = -1;
gint linenr = 0;
gint log_handler_encoder_id;
gboolean forking = TRUE;
sigjmp_buf environment; /* used for SIGSEGV handling */



/* Functions */

#ifdef STANDALONE
#ifdef __cplusplus
extern "C" {
void log_handler(const gchar *log_domain,GLogLevelFlags log_level,const gchar *message,gpointer user_data) {
    time_t Time = time(NULL);
    GString *date;
    GString *domain;
    GString *level;
    GString *logstring;
    GString *displaystring;
    gboolean quit;
    gint bytes = 0;
    gchar *last_char;

    /* Handle log_level, strip flags */
    log_level = (GLogLevelFlags)(log_level & 252);
    switch (log_level) {
        case G_LOG_LEVEL_ERROR:      level = g_string_new("Error");
                                     quit = TRUE;
                                     break;
        case G_LOG_LEVEL_CRITICAL:   level = g_string_new("Critical");
                                     quit = TRUE;
                                     break;
        case G_LOG_LEVEL_WARNING:    level = g_string_new("Warning");
                                     quit = FALSE;
                                     break;
        case G_LOG_LEVEL_MESSAGE:    level = g_string_new("Message");
                                     quit = FALSE;
                                     break;
        case G_LOG_LEVEL_INFO:       level = g_string_new("Info");
                                     quit = FALSE;
                                     break;
        case G_LOG_LEVEL_DEBUG:      level = g_string_new("Debug");
                                     quit = FALSE;
                                     break;
        default:                     level = g_string_new("Unknown");
                                     quit = FALSE;
                                     break;
    }

    /* Text formatting */
    while (level->len < 8) {
        level =  g_string_append(level," ");
    }

    /* Format date string */
    date = g_string_new(ctime(&Time));
    date = g_string_truncate(date,(date->len)-1);

    /* Store log entry */
    logstring = g_string_new("");
    g_string_sprintf(logstring,"%s\t| %i\t| %s | %s\n",date->str,linenr,level->str,message);
    displaystring = g_string_new("");
    g_string_sprintf(displaystring,"%s | %s\n",date->str,message);
    last_char = (gchar*)((glong)logstring->str + logstring->len - 2); /* we got 2x \n? */
    if (*last_char == 10) {
        logstring = g_string_truncate(logstring,logstring->len -1);
    }

    /* Print on STDOUT */
    g_print("%s",displaystring->str);

    /* Should we exit? */
    if (quit) {
       g_print("sleeping...");
       sleep(10);
       g_print(" done\n");
       exit(-1);
    }

    /* Clean & Exit */
    g_string_free(date,TRUE);
    g_string_free(level,TRUE);
    g_string_free(logstring,TRUE);
    g_string_free(displaystring,TRUE);
    return;
}
}
#endif
#endif

/* Handles signals */
void signal_handler(int sig) {

#ifndef STANDALONE
    gui_errorsignal(sig);
#else
    g_print("Got an error (%i), please report details of this bug",sig);
#endif

    if (sig == SIGSEGV) {
        /* Restore situation as it was before the segvault signal */
        siglongjmp(environment,1);
        printf("After SIGSEGV catching longjmp? Something is very wrong here...\n");
        fflush(0);
        signal(SIGSEGV,SIG_DFL); /* Reset segv signal handler, something has gone very wrong */
    }

    fflush(0);
    exit(0);
    return;
}

/* Hooks signals to internal handler */
void signal_hook(void) {

    /* Save environment */
    sigsetjmp(environment,1);

    if (sigsetjmp(environment,1) == 0) {
        /* Hook signals */
        signal(SIGTERM,signal_handler);
        signal(SIGHUP,signal_handler);
        signal(SIGABRT,signal_handler);
        signal(SIGTERM,signal_handler);
        //signal(,signal_handler);
        //signal(,signal_handler);
        //signal(,signal_handler);
        //signal(,signal_handler);
        //signal(SIGSEGV,signal_handler);
        //signal(SIGINT,signal_handler);
    } else {
        /* Got a SIGSEGV inside handler? */
        signal(SIGSEGV,SIG_DFL);
    }

    return;
}

int main(int argc, char** argv) {
    gint i;
    arg_parmS params;

#ifdef STANDALONE
    /* Init logging */
    gint log_handler_drip_id = g_log_set_handler (DRIP_LD, (GLogLevelFlags)(G_LOG_LEVEL_WARNING |G_LOG_LEVEL_MESSAGE | G_LOG_LEVEL_INFO | G_LOG_LEVEL_DEBUG | G_LOG_LEVEL_CRITICAL | G_LOG_LEVEL_ERROR | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION),log_handler, NULL);
#endif

    g_log(DRIP_LD,G_LOG_LEVEL_INFO,APPNAME " " VERSION " by Jarl van Katwijk <jarl@xs4all.nl>, based on work by Ulrich Hecht <uli@emulinks.de>");

    /* Check libavi version */
    if (GetAvifileVersion()!=AVIFILE_VERSION) {
        g_log(DRIP_LD,G_LOG_LEVEL_ERROR,"This binary was compiled for Avifile ver. %s, but the library is ver. %s. Aborting.",AVIFILE_VERSION,GetAvifileVersion());
    }
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Started dripencoder, pid %i",getpid());

    /* Init threading */
    g_thread_init(NULL);

    /* Pass params */
    params.argc = argc;
    params.argv = argv;
    encode((gpointer)&params);

    /* Clean & exit */
    exit(0);
    return(0);
}

