#include "mpeg3private.h"
#include "mpeg3protos.h"

#include <mntent.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

#define CONCAT_CHARACTER ':'

mpeg3_fs_t* mpeg3_new_fs(char *path)
{
	mpeg3_fs_t *fs = calloc(1, sizeof(mpeg3_fs_t));
	fs->css = mpeg3_new_css();
	strcpy(fs->path, path);
	return fs;
}

int mpeg3_delete_fs(mpeg3_fs_t *fs)
{
	mpeg3_delete_css(fs->css);
	free(fs);
	return 0;
}

int mpeg3_copy_fs(mpeg3_fs_t *dst, mpeg3_fs_t *src)
{
	strcpy(dst->path, src->path);
	dst->current_byte = 0;
	return 0;
}

off_t mpeg3io_get_total_bytes(mpeg3_fs_t *fs)
{
    off_t count = 0;
    int t;
    for(t=0; t < fs->numfiles; t++) {
        off_t size;
        fs->filepos[t] = count;
 
        mpeg3_get_keys(fs->css, fs->paths[t]);
        if(!(fs->fd = fopen64(fs->paths[t], "rb"))) {
               perror("mpeg3io_open_file(2)");
               return 0;
        }

	fseeko64(fs->fd, (off_t)0, SEEK_END);
	size = ftello64(fs->fd);
	fseeko64(fs->fd, (off_t)0, SEEK_SET);
        fclose(fs->fd);
        fs->fd = 0;
        count += size;
    }
    fs->filepos[t] = count;
    return count;
}

off_t mpeg3io_path_total_bytes(char *path)
{
	struct stat64 st;
	if(stat64(path, &st) < 0) return 0;
	return (off_t)st.st_size;
}

int mpeg3io_open_file(mpeg3_fs_t *fs)
{
/* Need to perform authentication before reading a single byte. */
	//mpeg3_get_keys(fs->css, fs->path);
        char*path;
        char*p;
        strcpy(fs->path_s, fs->path);
        path = fs->path_s;
        p = strchr(path, CONCAT_CHARACTER);
        fs->numfiles = 1;
        fs->paths[0] = path;
        while(p) {
            *p = 0;
            if(fs->numfiles >= MPEG3IO_MAX_FILE_CHUNKS)
                break;
            path = p + 1;
            fs->paths[fs->numfiles] = path;
            fs->numfiles++;
            p = strchr(path, CONCAT_CHARACTER);
        }


        fs->total_bytes = mpeg3io_get_total_bytes(fs);
       
        mpeg3_get_keys(fs->css, fs->paths[0]);
        if(!(fs->fd = fopen64(fs->paths[0], "rb"))) {
	    perror("mpeg3io_open_file(1)");
	    return 1;
	}

	if(!fs->total_bytes)
	{
		fclose(fs->fd);
		fs->fd = 0;
		return 1;
	}
	fs->current_byte = 0;
	fs->current_file = 0;
	return 0;
}

int mpeg3io_close_file(mpeg3_fs_t *fs)
{
	if(fs->fd) fclose(fs->fd);
	fs->fd = 0;
	return 0;
}

int mpeg3io_read_data(unsigned char *buffer, off_t bytes, mpeg3_fs_t *fs) {
       if(!bytes) {
           return 0;
       }

       if(fs->current_byte + bytes > fs->filepos[fs->current_file + 1]) {
           off_t firstbytes = fs->filepos[fs->current_file + 1] - fs->current_byte;
           if(firstbytes) {
               if(!fread(buffer, 1, firstbytes, fs->fd)) {
                   return 1;
               }
           }
           
           if(fs->current_file + 1 == fs->numfiles) {
               fs->current_byte = fs->total_bytes;
               /* the return value here should be 1, as we didn't read
                * the full bufferlength. However, libmpeg3-1.2.3 expects
                * to only get an error for _empty_ buffers. */
               return 0; 
           }
           fclose (fs->fd);
           fs->current_file ++;
           mpeg3_get_keys(fs->css, fs->paths[fs->current_file]);
           fs->fd = fopen64(fs->paths[fs->current_file], "rb");
           if(!(fs->fd)) {
                   perror("mpeg3io_open_file(3)");
                   return 1;
           }
   
           fs->current_byte = fs->filepos[fs->current_file];
           bytes -= firstbytes;
           buffer += firstbytes;
       }
       
       if(!fread(buffer, 1, bytes, fs->fd)) {
           return 1;
       }
	fs->current_byte += bytes;
	return 0;
}

int mpeg3io_device(char *path, char *device)
{
	struct stat64 file_st, device_st;
	struct mntent *mnt;
	FILE *fp;

	if(stat64(path, &file_st) < 0)
	{
		perror("mpeg3io_device");
		return 1;
	}

	fp = setmntent(MOUNTED, "r");
	while(fp && (mnt = getmntent(fp)))
	{
		if(stat64(mnt->mnt_fsname, &device_st) < 0) continue;
		if(device_st.st_rdev == file_st.st_dev)
		{
			strncpy(device, mnt->mnt_fsname, MPEG3_STRLEN);
			break;
		}
	}
	endmntent(fp);

	return 0;
}

int mpeg3io_seek(mpeg3_fs_t *fs, off_t byte)
{
       int s, t=-1, ret;
       if(byte >= fs->total_bytes) {
           return -1;
       }

	fs->current_byte = byte;
       for(s = 0; s < fs->numfiles; s++) {
           if(byte < fs->filepos[s + 1]) {
               t=s;break;
           }
       }
       if(t<0) {
           perror("internal error: mpeg3io_seek");
           return -1;
       }
       if(t != fs->current_file) {
           fclose(fs->fd);
           fs->current_file = t;
           mpeg3_get_keys(fs->css, fs->paths[fs->current_file]);
           fs->fd = fopen64(fs->paths[fs->current_file], "rb");
           if(!(fs->fd))
           {
                   perror("mpeg3io_open_file(4)");
                   return -1;
           }
       }
       ret = fseeko64(fs->fd, byte - fs->filepos[fs->current_file], SEEK_SET);
       return ret;

}

int mpeg3io_seek_relative(mpeg3_fs_t *fs, off_t bytes)
{
       off_t pos = fs->current_byte + bytes;
       return mpeg3io_seek(fs, pos);
}

void mpeg3io_complete_path(char *complete_path, char *path)
{
	if(path[0] != '/')
	{
		char current_dir[MPEG3_STRLEN];
		getcwd(current_dir, MPEG3_STRLEN);
		sprintf(complete_path, "%s/%s", current_dir, path);
	}
	else
		strcpy(complete_path, path);
}

void mpeg3io_get_directory(char *directory, char *path)
{
	char *ptr = strrchr(path, '/');
	if(ptr)
	{
		int i;
		for(i = 0; i < ptr - path; i++)
		{
			directory[i] = path[i];
		}
		directory[i] = 0;
	}
}

void mpeg3io_get_filename(char *filename, char *path)
{
	char *ptr = strrchr(path, '/');
	if(!ptr) 
		ptr = path;
	else
		ptr++;

	strcpy(filename, ptr);
}

void mpeg3io_joinpath(char *title_path, char *directory, char *new_filename)
{
	sprintf(title_path, "%s/%s", directory, new_filename);
}

off_t mpeg3io_tell(mpeg3_fs_t *fs) {
	int *segfault;
	printf("mpeg3io_tell\n");fflush(NULL);
	*segfault = 0xDEAD;
        return (off_t)fs->current_byte;
}

unsigned int mpeg3io_read_char(mpeg3_fs_t *fs)
{
       unsigned char byte;
       mpeg3io_read_data(&byte, 1, fs);
       return byte;
}

unsigned int mpeg3io_read_int32(mpeg3_fs_t *fs)
{
        int a, b, c, d;
        unsigned int result;
/* Do not fread.  This breaks byte ordering. */
        a = (unsigned char)mpeg3io_read_char(fs);
        b = (unsigned char)mpeg3io_read_char(fs);
        c = (unsigned char)mpeg3io_read_char(fs);
        d = (unsigned char)mpeg3io_read_char(fs);
        result = ((int)a << 24) | ((int)b << 16) | ((int)c << 8) | ((int)d);
        return result;
}


