
/*
 * OVERLAY A SUBPICTURE OVER ANOTHER IMAGE (USING 3BytePerPixel)
 *
 * Copyright (C) 2000  Thomas Mirlacher
 * Patched to 24bits and reverse vertical by Jarl van Katwijk
 * 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * The author may be reached as dent@cosy.sbg.ac.at, or
 * Thomas Mirlacher, Jakob-Haringerstr. 2, A-5020 Salzburg,
 * Austria
 *
 *------------------------------------------------------------
 *
 */


#include <sys/types.h>
#include <glib.h>
#include "spu.h"

u_int *overlay (guint *background,gint xb,gint yb,gint ob,struct spu_img_struct *spu,gfloat o1) {
/* Render 8bits subpicture over 24bits main movie frame */
/* TODO: implement variable background sizes */
    glong x, y;
    gulong inv_subp,trn_back;
    guint8 new_pixel;
    gint i;
    guchar *clr_background;
    guchar *clr_subpicture;
    guint clr_subtrans,inv_subtrans;
    glong background_x = (glong)background + ((glong)(spu->x))*4;
    /* Calculate offset of subpictures on framebuffer */
    glong background_top = (glong)background + ((yb+ob-(yb-(spu->y+spu->height)))*xb*3+(spu->x)*3); // heigth correction, all subs rendered on botton of frame
    guchar *spu_data_ptr = (guchar *) spu->data;
    gboolean printed = FALSE;

    /* Check */
    if ((spu->width <= 0) || (spu->height <= 0) ||
        (spu->width > 1024) || (spu->height > 1024)) {
        printf ("width || height out of range.\n");
        return NULL;
    }

    /* Render subpicture over background */
    for (y=(spu->y);y<(spu->height+spu->y-1); y++) {
        clr_background = (guchar*)(background_top - y*3*xb);
        if ((glong)clr_background<background_top && (glong)clr_background>=(glong)background) {
            for (x=spu->x;x<(spu->width+spu->x); x++) {
                clr_subtrans = (*spu_data_ptr)&0xf0;
                switch (clr_subtrans) {
                case 0:
                    /* Total transparency, do nothing */
                    *clr_background++;
                    *clr_background++;
                    *clr_background++;
                    break;
                case 240:
                    /* No transparency, just overwrite */
                    clr_subpicture = (gchar*)&spu_clut[(*spu_data_ptr)&15];
                    *clr_background = (guchar)(*clr_subpicture);
                    clr_background++;clr_subpicture++;
                    *clr_background = (guchar)(*clr_subpicture);
                    clr_background++;clr_subpicture++;
                    *clr_background = (guchar)(*clr_subpicture);
                    clr_background++;clr_subpicture++;
                    break;
                default:
                    /* Apply subpicture pixel with transparency */
                    clr_subpicture = (gchar*)&spu_clut[(*spu_data_ptr)&15];
                    inv_subtrans = clr_subtrans;
                    clr_subtrans = 256-inv_subtrans;
    
                    inv_subp = (*clr_subpicture) * inv_subtrans;
                    trn_back = (*clr_background) * clr_subtrans;
                    new_pixel = (inv_subp + trn_back)>>8;
                    *clr_background = (guchar)new_pixel;
                    *clr_background++;*clr_subpicture++;

                    inv_subp = (*clr_subpicture) * inv_subtrans;
                    trn_back = (*clr_background) * clr_subtrans;
                    new_pixel = (inv_subp + trn_back)>>8;
                    *clr_background = (guchar)new_pixel;
                    *clr_background++;*clr_subpicture++;

                    inv_subp = (*clr_subpicture) * inv_subtrans;
                    trn_back = (*clr_background) * clr_subtrans;
                    new_pixel = (inv_subp + trn_back)>>8;
                    *clr_background = (guchar)new_pixel;
                    *clr_background++;*clr_subpicture++;
                    break;
                }
                spu_data_ptr++;
            }
        }
    }

    return background;
}


